;;; font-lock-ignore.el --- Selectively remove font-lock keywords  -*- lexical-binding: t; -*-

;; Copyright (C) 2022  Augusto Stoffel

;; Author: Augusto Stoffel <arstoffel@gmail.com>
;; Keywords: languages, faces

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; 

;;; Code:

(require 'font-lock)
(require 'seq)

(defvar font-lock-ignore-rules nil
  "Rules to selectively remove font-lock keywords.
This is a list of rules of the form.

  (MODE [!] RULE ...)

where:

 - MODE is a symbol, say a major or minor mode.  The subsequent
   rules apply if the current major mode is derived from MODE or
   MODE is bound and true as a variable.

 - RULE can be:
   - A symbol, say a face name.  It matches any font-lock rule
     mentioning that symbol anywhere.
   - A string.  It matches any font-lock keyword where a symbol or
     string matches the string, interpreted as a regexp.
   - A sexp (matches STRING).  It matches any font-lock keyword
     defined by a regexp that matches STRING.
   - The symbol `*'.  Matches everything.

Any font lock keyword that matches a rule is excluded.  Rules
preceded by `!' have the opposite effect: they ensure a keyword
will be applied even if a previous rule excluded it.")

(defun font-lock-ignore--test-one (rule keyword)
  "Test whether font-lock KEYWORD matches a RULE.
See `font-lock-ignore-rules' for the possible rules."
  (pcase-exhaustive rule
    ('* t)
    (`(matches ,string) (when (stringp (car keyword))
                          (string-match-p (concat "\\`\\(?:" (car keyword) "\\)")
                                          string)))
    ((pred stringp) (seq-some (lambda (obj)
                                (when-let ((str (cond
                                                 ((stringp obj) obj)
                                                 ((symbolp obj) (symbol-name obj)))))
                                  (string-match-p rule str)))
                             (flatten-tree keyword)))
    ((pred symbolp) (memq rule (flatten-tree keyword)))))

(defun font-lock-ignore--test (rules keyword)
  "Test whether font-lock KEYWORD matches the RULES.
Each element of RULES is a RULE in the sense of
`font-lock-ignore-rules', possibly preceded by an exclamation
mark."
  (let ((keep t))
    (while rules
      (pcase-exhaustive rules
        (`(! ,rule . ,rest)
         (setq rules rest
               keep (or keep (font-lock-ignore--test-one rule keyword))))
        (`(,rule . ,rest)
         (setq rules rest
               keep (and keep (not (font-lock-ignore--test-one rule keyword)))))))
    keep))

(defun font-lock-ignore--filter (keywords)
  "Filter a list of compiled font lock KEYWORDS.
KEYWORDS is supposed to be compiled (so its car is the symbol t).
The before-compilation version of the keywords, which is expected
on the cadr of KEYWORDS, remains unchanged."
  (let ((rules (seq-mapcat (pcase-lambda (`(,mode . ,rules))
                             (when (or (and (boundp mode) mode)
                                       (derived-mode-p mode))
                               rules))
                           font-lock-ignore-rules)))
    (append (seq-take keywords 2)
            (seq-filter (lambda (kw) (font-lock-ignore--test rules kw))
                        (cddr keywords)))))

;;;###autoload
(define-minor-mode font-lock-ignore-mode
  "Selectively remove font lock."
  :global t
  (if font-lock-ignore-mode
      (advice-add 'font-lock-compile-keywords :filter-return #'font-lock-ignore--filter)
    (advice-remove 'font-lock-compile-keywords #'font-lock-ignore--filter))
  (dolist (buffer (buffer-list))
    (with-current-buffer buffer
      (when font-lock-set-defaults
        (when (eq t (car font-lock-keywords))
          (setq font-lock-keywords (font-lock-compile-keywords
                                    (cadr font-lock-keywords)))
          (font-lock-flush))))))

(defun font-lock-ignore-inspect ()
  "Display all font lock keyword for the current buffer."
  (interactive)
  (let ((original (cadr font-lock-keywords))
        (compiled (cddr font-lock-keywords))
        (buffer (get-buffer-create "*Font-lock Keywords*")))
    (with-current-buffer buffer
      (erase-buffer)
      (dolist (keyword original)
        (let* ((matcher (or (car-safe keyword) keyword))
               (comp (assoc matcher compiled)))
          (insert "Matcher: " (pp-to-string matcher))
          (unless (eq (char-before) ?\n) (insert ?\n))
          (insert "Highlight: " (if comp (pp-to-string (cdr-safe keyword)) "ignored")))
        (insert (if (eq (char-before) ?\n) ?\n "\n\n"))))
    (display-buffer buffer)))

(provide 'font-lock-ignore)
;;; font-lock-ignore.el ends here
